@@ -71,7 +71,8 @@ def _is_daemon():
 
 
 def _is_sighup_and_daemon(signo):
-    if not (SignalHandler().is_sighup_supported and signo == signal.SIGHUP):
+    if not (SignalHandler().is_signal_supported('SIGHUP') and
+            signo == signal.SIGHUP):
         # Avoid checking if we are a daemon, because the signal isn't
         # SIGHUP.
         return False
@@ -135,7 +136,6 @@ class SignalHandler(object):
         self.signals_to_name = dict(
             (sigval, name)
             for (name, sigval) in self._signals_by_name.items())
-        self.is_sighup_supported = 'SIGHUP' in self._signals_by_name
         self._signal_handlers = collections.defaultdict(set)
         self.clear()
 
@@ -149,16 +149,35 @@ class SignalHandler(object):
             self.add_handler(sig, handler)
 
     def add_handler(self, sig, handler):
-        if sig == "SIGHUP" and not self.is_sighup_supported:
+        if not self.is_signal_supported(sig):
             return
         signo = self._signals_by_name[sig]
         self._signal_handlers[signo].add(handler)
-        signal.signal(signo, self._handle_signals)
+        signal.signal(signo, self._handle_signal)
 
-    def _handle_signals(self, signo, frame):
+    def _handle_signal(self, signo, frame):
+        # This method can be called anytime, even between two Python
+        # instructions. It's scheduled by the C signal handler of Python using
+        # Py_AddPendingCall().
+        #
+        # We only do one thing: schedule a call to _handle_signal_cb() later.
+        # eventlet.spawn() is not signal-safe: _handle_signal() can be called
+        # during a call to eventlet.spawn(). This case is supported, it is
+        # ok to schedule multiple calls to _handle_signal() with the same
+        # signal number.
+        #
+        # To call to _handle_signal_cb() is delayed to avoid reentrant calls to
+        # _handle_signal_cb(). It avoids race conditions like reentrant call to
+        # clear(): clear() is not reentrant (bug #1538204).
+        eventlet.spawn(self._handle_signal_cb, signo, frame)
+
+    def _handle_signal_cb(self, signo, frame):
         for handler in self._signal_handlers[signo]:
             handler(signo, frame)
 
+    def is_signal_supported(self, sig_name):
+        return sig_name in self._signals_by_name
+
 
 class Launcher(object):
     """Launch one or more services and wait for them to complete."""
@@ -227,25 +246,36 @@ class ServiceLauncher(Launcher):
         :param conf: an instance of ConfigOpts
         """
         super(ServiceLauncher, self).__init__(conf)
+        self.signal_handler = SignalHandler()
 
-    def _handle_signal(self, signo, frame):
-        """Set signal handlers and raise an exception.
+    def _graceful_shutdown(self, *args):
+        self.signal_handler.clear()
+        if (self.conf.graceful_shutdown_timeout and
+                self.signal_handler.is_signal_supported('SIGALRM')):
+            signal.alarm(self.conf.graceful_shutdown_timeout)
+        self.stop()
 
-        :param signo: signal number
-        :param frame: current stack frame
-        :raises SignalExit
-        """
-        # Allow the process to be killed again and die from natural causes
-        SignalHandler().clear()
-        raise SignalExit(signo)
+    def _reload_service(self, *args):
+        self.signal_handler.clear()
+        raise SignalExit(signal.SIGHUP)
+
+    def _fast_exit(self, *args):
+        LOG.info(_LI('Caught SIGINT signal, instantaneous exiting'))
+        os._exit(1)
+
+    def _on_timeout_exit(self, *args):
+        LOG.info(_LI('Graceful shutdown timeout exceeded, '
+                     'instantaneous exiting'))
+        os._exit(1)
 
     def handle_signal(self):
         """Set self._handle_signal as a signal handler."""
-        SignalHandler().add_handlers(
-            ('SIGTERM', 'SIGHUP', 'SIGINT'),
-            self._handle_signal)
+        self.signal_handler.add_handler('SIGTERM', self._graceful_shutdown)
+        self.signal_handler.add_handler('SIGINT', self._fast_exit)
+        self.signal_handler.add_handler('SIGHUP', self._reload_service)
+        self.signal_handler.add_handler('SIGALRM', self._on_timeout_exit)
 
-    def _wait_for_exit_or_signal(self, ready_callback=None):
+    def _wait_for_exit_or_signal(self):
         status = None
         signo = 0
 
@@ -254,35 +284,36 @@ class ServiceLauncher(Launcher):
             self.conf.log_opt_values(LOG, logging.DEBUG)
 
         try:
-            if ready_callback:
-                ready_callback()
             super(ServiceLauncher, self).wait()
         except SignalExit as exc:
-            signame = SignalHandler().signals_to_name[exc.signo]
+            signame = self.signal_handler.signals_to_name[exc.signo]
             LOG.info(_LI('Caught %s, exiting'), signame)
             status = exc.code
             signo = exc.signo
         except SystemExit as exc:
-            status = exc.code
-        finally:
             self.stop()
-
+            status = exc.code
+        except Exception:
+            self.stop()
         return status, signo
 
-    def wait(self, ready_callback=None):
+    def wait(self):
         """Wait for a service to terminate and restart it on SIGHUP.
 
         :returns: termination status
         """
         systemd.notify_once()
-        SignalHandler().clear()
+        self.signal_handler.clear()
         while True:
             self.handle_signal()
-            status, signo = self._wait_for_exit_or_signal(ready_callback)
+            status, signo = self._wait_for_exit_or_signal()
             if not _is_sighup_and_daemon(signo):
-                return status
+                break
             self.restart()
 
+        super(ServiceLauncher, self).wait()
+        return status
+
 
 class ServiceWrapper(object):
     def __init__(self, service, workers):
@@ -319,6 +350,7 @@ class ProcessLauncher(object):
         self.signal_handler.add_handlers(('SIGTERM', 'SIGHUP'),
                                          self._handle_signal)
         self.signal_handler.add_handler('SIGINT', self._fast_exit)
+        self.signal_handler.add_handler('SIGALRM', self._on_alarm_exit)
 
     def _handle_signal(self, signo, frame):
         """Set signal handlers.
@@ -336,6 +368,11 @@ class ProcessLauncher(object):
         LOG.info(_LI('Caught SIGINT signal, instantaneous exiting'))
         os._exit(1)
 
+    def _on_alarm_exit(self, signo, frame):
+        LOG.info(_LI('Graceful shutdown timeout exceeded, '
+                     'instantaneous exiting'))
+        os._exit(1)
+
     def _pipe_watcher(self):
         # This will block until the write end is closed when the parent
         # dies unexpectedly
@@ -352,11 +389,11 @@ class ProcessLauncher(object):
         # Setup child signal handlers differently
 
         def _sigterm(*args):
-            SignalHandler().clear()
+            self.signal_handler.clear()
             self.launcher.stop()
 
         def _sighup(*args):
-            SignalHandler().clear()
+            self.signal_handler.clear()
             raise SignalExit(signal.SIGHUP)
 
         self.signal_handler.clear()
@@ -435,7 +472,7 @@ class ProcessLauncher(object):
 
             os._exit(status)
 
-        LOG.info(_LI('Started child %d'), pid)
+        LOG.debug('Started child %d', pid)
 
         wrap.children.add(pid)
         self.children[pid] = wrap
@@ -524,13 +561,19 @@ class ProcessLauncher(object):
                     service.reset()
 
                 for pid in self.children:
-                    os.kill(pid, signal.SIGHUP)
+                    os.kill(pid, signal.SIGTERM)
 
                 self.running = True
                 self.sigcaught = None
         except eventlet.greenlet.GreenletExit:
             LOG.info(_LI("Wait called after thread killed. Cleaning up."))
 
+        # if we are here it means that we are trying to do graceful shutdown.
+        # add alarm watching that graceful_shutdown_timeout is not exceeded
+        if (self.conf.graceful_shutdown_timeout and
+                self.signal_handler.is_signal_supported('SIGALRM')):
+            signal.alarm(self.conf.graceful_shutdown_timeout)
+
         self.stop()
 
     def stop(self):
@@ -563,13 +606,8 @@ class Service(ServiceBase):
     def __init__(self, threads=1000):
         self.tg = threadgroup.ThreadGroup(threads)
 
-        # signal that the service is done shutting itself down:
-        self._done = event.Event()
-
     def reset(self):
         """Reset a service in case it received a SIGHUP."""
-        # NOTE(Fengqian): docs for Event.reset() recommend against using it
-        self._done = event.Event()
 
     def start(self):
         """Start a service."""
@@ -581,14 +619,10 @@ class Service(ServiceBase):
                or terminate them instantly
         """
         self.tg.stop(graceful)
-        self.tg.wait()
-        # Signal that service cleanup is done:
-        if not self._done.ready():
-            self._done.send()
 
     def wait(self):
         """Wait for a service to shut down."""
-        self._done.wait()
+        self.tg.wait()
 
 
 class Services(object):
@@ -642,8 +676,13 @@ class Services(object):
         :returns: None
 
         """
-        service.start()
-        done.wait()
+        try:
+            service.start()
+        except Exception:
+            LOG.exception(_LE('Error starting thread.'))
+            raise SystemExit(1)
+        else:
+            done.wait()
 
 
 def launch(conf, service, workers=1):
@@ -655,6 +694,10 @@ def launch(conf, service, workers=1):
     :param workers: a number of processes in which a service will be running
     :returns: instance of a launcher that was used to launch the service
     """
+
+    if workers is not None and workers <= 0:
+        raise ValueError("Number of workers should be positive!")
+
     if workers is None or workers == 1:
         launcher = ServiceLauncher(conf)
         launcher.launch_service(service)
