@@ -20,15 +20,11 @@ Unit Tests for service class
 
 from __future__ import print_function
 
-import threading
-
 import logging
 import multiprocessing
 import os
 import signal
 import socket
-import subprocess
-import sys
 import time
 import traceback
 
@@ -36,10 +32,10 @@ import eventlet
 from eventlet import event
 import mock
 from oslotest import base as test_base
-from six.moves import queue
 
 from oslo_service import service
 from oslo_service.tests import base
+from oslo_service.tests import eventlet_service
 
 
 LOG = logging.getLogger(__name__)
@@ -59,19 +55,37 @@ class ServiceManagerTestCase(test_base.BaseTestCase):
 
 
 class ServiceWithTimer(service.Service):
+    def __init__(self, ready_event=None):
+        super(ServiceWithTimer, self).__init__()
+        self.ready_event = ready_event
+
     def start(self):
         super(ServiceWithTimer, self).start()
         self.timer_fired = 0
         self.tg.add_timer(1, self.timer_expired)
 
+    def wait(self):
+        if self.ready_event:
+            self.ready_event.set()
+        super(ServiceWithTimer, self).wait()
+
     def timer_expired(self):
         self.timer_fired = self.timer_fired + 1
 
 
+class ServiceCrashOnStart(ServiceWithTimer):
+    def start(self):
+        super(ServiceCrashOnStart, self).start()
+        raise ValueError
+
+
 class ServiceTestBase(base.ServiceBaseTestCase):
     """A base class for ServiceLauncherTest and ServiceRestartTest."""
 
-    def _spawn_service(self, workers=1, *args, **kwargs):
+    def _spawn_service(self,
+                       workers=1,
+                       service_to_launch=ServiceWithTimer,
+                       *args, **kwargs):
         self.workers = workers
         pid = os.fork()
         if pid == 0:
@@ -85,9 +99,9 @@ class ServiceTestBase(base.ServiceBaseTestCase):
             # os._exit() which doesn't have this problem.
             status = 0
             try:
-                serv = ServiceWithTimer()
+                serv = service_to_launch(*args, **kwargs)
                 launcher = service.launch(self.conf, serv, workers=workers)
-                launcher.wait(*args, **kwargs)
+                status = launcher.wait()
             except SystemExit as exc:
                 status = exc.code
             except BaseException:
@@ -98,7 +112,7 @@ class ServiceTestBase(base.ServiceBaseTestCase):
                     print("Couldn't print traceback")
                 status = 2
             # Really exit
-            os._exit(status)
+            os._exit(status or 0)
         return pid
 
     def _wait(self, cond, timeout):
@@ -113,8 +127,8 @@ class ServiceTestBase(base.ServiceBaseTestCase):
         # NOTE(markmc): ConfigOpts.log_opt_values() uses CONF.config-file
         self.conf(args=[], default_config_files=[])
         self.addCleanup(self.conf.reset)
-        self.addCleanup(self.conf.reset)
         self.addCleanup(self._reap_pid)
+        self.pid = 0
 
     def _reap_pid(self):
         if self.pid:
@@ -197,6 +211,12 @@ class ServiceLauncherTest(ServiceTestBase):
         self.assertTrue(os.WIFEXITED(status))
         self.assertEqual(os.WEXITSTATUS(status), 0)
 
+    def test_crashed_service(self):
+        self.pid = self._spawn_service(service_to_launch=ServiceCrashOnStart)
+        status = self._reap_test()
+        self.assertTrue(os.WIFEXITED(status))
+        self.assertEqual(os.WEXITSTATUS(status), 1)
+
     def test_child_signal_sighup(self):
         start_workers = self._spawn()
 
@@ -215,15 +235,16 @@ class ServiceLauncherTest(ServiceTestBase):
         start_workers = self._spawn()
 
         os.kill(self.pid, signal.SIGHUP)
-        # Wait at most 5 seconds to respawn a worker
-        cond = lambda: start_workers == self._get_workers()
-        timeout = 5
-        self._wait(cond, timeout)
 
-        # Make sure worker pids match
-        end_workers = self._get_workers()
-        LOG.info('workers: %r' % end_workers)
-        self.assertEqual(start_workers, end_workers)
+        def cond():
+            workers = self._get_workers()
+            return (len(workers) == len(start_workers) and
+                    not set(start_workers).intersection(workers))
+
+        # Wait at most 5 seconds to respawn a worker
+        timeout = 10
+        self._wait(cond, timeout)
+        self.assertTrue(cond())
 
 
 class ServiceRestartTest(ServiceTestBase):
@@ -231,7 +252,7 @@ class ServiceRestartTest(ServiceTestBase):
     def _spawn(self):
         ready_event = multiprocessing.Event()
         self.pid = self._spawn_service(workers=1,
-                                       ready_callback=ready_event.set)
+                                       ready_event=ready_event)
         return ready_event
 
     def test_service_restart(self):
@@ -285,7 +306,6 @@ class LauncherTest(base.ServiceBaseTestCase):
 
         launcher.stop()
         self.assertTrue(svc.cleaned_up)
-        self.assertTrue(svc._done.ready())
 
         # make sure stop can be called more than once.  (i.e. play nice with
         # unit test fixtures in nova bug #1199315)
@@ -303,8 +323,15 @@ class LauncherTest(base.ServiceBaseTestCase):
     def test_launch_one_worker(self):
         self._test_launch_single(1)
 
+    def test_launch_invalid_workers_number(self):
+        svc = service.Service()
+        for num_workers in [0, -1]:
+            self.assertRaises(ValueError, service.launch, self.conf,
+                              svc, num_workers)
+
+    @mock.patch('signal.alarm')
     @mock.patch('oslo_service.service.ProcessLauncher.launch_service')
-    def test_multiple_worker(self, mock_launch):
+    def test_multiple_worker(self, mock_launch, alarm_mock):
         svc = service.Service()
         service.launch(self.conf, svc, workers=3)
         mock_launch.assert_called_with(svc, workers=3)
@@ -315,19 +342,23 @@ class LauncherTest(base.ServiceBaseTestCase):
         svc = mock.Mock()
         self.assertRaises(TypeError, service.launch, self.conf, svc)
 
+    @mock.patch('signal.alarm')
     @mock.patch("oslo_service.service.Services.add")
     @mock.patch("oslo_service.eventlet_backdoor.initialize_if_enabled")
     def test_check_service_base(self, initialize_if_enabled_mock,
-                                services_mock):
+                                services_mock,
+                                alarm_mock):
         initialize_if_enabled_mock.return_value = None
         launcher = service.Launcher(self.conf)
         serv = _Service()
         launcher.launch_service(serv)
 
+    @mock.patch('signal.alarm')
     @mock.patch("oslo_service.service.Services.add")
     @mock.patch("oslo_service.eventlet_backdoor.initialize_if_enabled")
     def test_check_service_base_fails(self, initialize_if_enabled_mock,
-                                      services_mock):
+                                      services_mock,
+                                      alarm_mock):
         initialize_if_enabled_mock.return_value = None
         launcher = service.Launcher(self.conf)
 
@@ -340,8 +371,9 @@ class LauncherTest(base.ServiceBaseTestCase):
 
 class ProcessLauncherTest(base.ServiceBaseTestCase):
 
+    @mock.patch('signal.alarm')
     @mock.patch("signal.signal")
-    def test_stop(self, signal_mock):
+    def test_stop(self, signal_mock, alarm_mock):
         signal_mock.SIGTERM = 15
         launcher = service.ProcessLauncher(self.conf)
         self.assertTrue(launcher.running)
@@ -370,7 +402,7 @@ class ProcessLauncherTest(base.ServiceBaseTestCase):
                          mock_kill.mock_calls)
         mock_service_stop.assert_called_once_with()
 
-    def test__handle_signals(self):
+    def test__handle_signal(self):
         signal_handler = service.SignalHandler()
         signal_handler.clear()
         self.assertEqual(0,
@@ -380,11 +412,14 @@ class ProcessLauncherTest(base.ServiceBaseTestCase):
         signal_handler.add_handler('SIGTERM', call_2)
         self.assertEqual(2,
                          len(signal_handler._signal_handlers[signal.SIGTERM]))
-        signal_handler._handle_signals(signal.SIGTERM, 'test')
+        signal_handler._handle_signal(signal.SIGTERM, 'test')
+        # execute pending eventlet callbacks
+        time.sleep(0)
         for m in signal_handler._signal_handlers[signal.SIGTERM]:
             m.assert_called_once_with(signal.SIGTERM, 'test')
         signal_handler.clear()
 
+    @mock.patch('signal.alarm')
     @mock.patch("os.kill")
     @mock.patch("oslo_service.service.ProcessLauncher.stop")
     @mock.patch("oslo_service.service.ProcessLauncher._respawn_children")
@@ -401,7 +436,8 @@ class ProcessLauncherTest(base.ServiceBaseTestCase):
                                           handle_signal_mock,
                                           respawn_children_mock,
                                           stop_mock,
-                                          kill_mock):
+                                          kill_mock,
+                                          alarm_mock):
         is_sighup_and_daemon_mock.return_value = True
         respawn_children_mock.side_effect = [None,
                                              eventlet.greenlet.GreenletExit()]
@@ -470,7 +506,7 @@ def exercise_graceful_test_service(sleep_amount, time_to_wait, graceful):
 class ServiceTest(test_base.BaseTestCase):
     def test_graceful_stop(self):
         # Here we wait long enough for the task to gracefully finish.
-        self.assertEqual(None, exercise_graceful_test_service(1, 2, True))
+        self.assertIsNone(exercise_graceful_test_service(1, 2, True))
 
     def test_ungraceful_stop(self):
         # Here we stop ungracefully, and will never see the task finish.
@@ -478,39 +514,21 @@ class ServiceTest(test_base.BaseTestCase):
                          exercise_graceful_test_service(1, 2, False))
 
 
-class EventletServerTest(test_base.BaseTestCase):
+class EventletServerProcessLauncherTest(base.ServiceBaseTestCase):
+    def setUp(self):
+        super(EventletServerProcessLauncherTest, self).setUp()
+        self.conf(args=[], default_config_files=[])
+        self.addCleanup(self.conf.reset)
+        self.workers = 3
+
     def run_server(self):
+        queue = multiprocessing.Queue()
+        proc = multiprocessing.Process(target=eventlet_service.run,
+                                       args=(queue,),
+                                       kwargs={'workers': self.workers})
+        proc.start()
 
-        server_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
-                                   'eventlet_service.py')
-
-        # Start up an eventlet server.
-        server = subprocess.Popen([sys.executable, server_path],
-                                  stdout=subprocess.PIPE,
-                                  stderr=subprocess.PIPE,
-                                  bufsize=1000,
-                                  close_fds=True)
-
-        def enqueue_output(f, q):
-            while True:
-                line = f.readline()
-                if not line:
-                    break
-                q.put(line)
-            f.close()
-
-        # Start a thread to read stderr so the app doesn't block.
-        err_q = queue.Queue()
-        err_t = threading.Thread(target=enqueue_output,
-                                 args=(server.stderr, err_q))
-        err_t.daemon = True
-        err_t.start()
-
-        # The server's line of output is the port it picked.
-        port_str = server.stdout.readline()
-        port = int(port_str)
-
-        # connect to the server.
+        port = queue.get()
         conn = socket.create_connection(('127.0.0.1', port))
 
         # NOTE(blk-u): The sleep shouldn't be necessary. There must be a bug in
@@ -518,38 +536,55 @@ class EventletServerTest(test_base.BaseTestCase):
         # server or signal handlers.
         time.sleep(1)
 
-        return (server, conn)
+        return (proc, conn)
 
     def test_shuts_down_on_sigint_when_client_connected(self):
-        server, conn = self.run_server()
+        proc, conn = self.run_server()
 
         # check that server is live
-        self.assertIsNone(server.poll())
+        self.assertTrue(proc.is_alive())
 
         # send SIGINT to the server and wait for it to exit while client still
         # connected.
-        server.send_signal(signal.SIGINT)
-        server.wait()
+        os.kill(proc.pid, signal.SIGINT)
+        proc.join()
         conn.close()
 
     def test_graceful_shuts_down_on_sigterm_when_client_connected(self):
-        server, conn = self.run_server()
+        self.config(graceful_shutdown_timeout=7)
+        proc, conn = self.run_server()
 
         # send SIGTERM to the server and wait for it to exit while client still
         # connected.
-        server.send_signal(signal.SIGTERM)
+        os.kill(proc.pid, signal.SIGTERM)
 
-        server_wait_thread = threading.Thread(
-            target=lambda server: server.wait(), args=(server,))
-        server_wait_thread.start()
-
-        # server with graceful shutdown must wait forewer
-        # for closing connection by client
-        # but for test 3 seconds is enough
+        # server with graceful shutdown must wait forewer if
+        # option graceful_shutdown_timeout is not specified.
+        # we can not wait forever ... so 3 seconds are enough
         time.sleep(3)
 
-        self.assertEqual(True, server_wait_thread.is_alive())
+        self.assertTrue(proc.is_alive())
 
         conn.close()
+        proc.join()
 
-        server_wait_thread.join()
+    def test_graceful_stop_with_exceeded_graceful_shutdown_timeout(self):
+        # Server must exit if graceful_shutdown_timeout exceeded
+        graceful_shutdown_timeout = 4
+        self.config(graceful_shutdown_timeout=graceful_shutdown_timeout)
+        proc, conn = self.run_server()
+
+        time_before = time.time()
+        os.kill(proc.pid, signal.SIGTERM)
+        self.assertTrue(proc.is_alive())
+        proc.join()
+        self.assertFalse(proc.is_alive())
+        time_after = time.time()
+
+        self.assertTrue(time_after - time_before > graceful_shutdown_timeout)
+
+
+class EventletServerServiceLauncherTest(EventletServerProcessLauncherTest):
+    def setUp(self):
+        super(EventletServerServiceLauncherTest, self).setUp()
+        self.workers = 1
