@@ -33,6 +33,7 @@ import webob.exc
 
 from oslo_log import log as logging
 from oslo_service import _options
+from oslo_service import service
 from oslo_service import sslutils
 from oslo_service._i18n import _, _LE, _LI
 
@@ -45,28 +46,47 @@ def list_opts():
     return [(None, copy.deepcopy(_options.wsgi_opts))]
 
 
+def register_opts(conf):
+    """Registers WSGI config options."""
+    return conf.register_opts(_options.wsgi_opts)
+
+
 class InvalidInput(Exception):
     message = _("Invalid input received: "
                 "Unexpected argument for periodic task creation: %(arg)s.")
 
 
-class Server(object):
+class Server(service.ServiceBase):
     """Server class to manage a WSGI server, serving a WSGI application."""
 
-    def __init__(self, conf, name, app, host='0.0.0.0', port=0, pool_size=None,
-                 protocol=eventlet.wsgi.HttpProtocol, backlog=128,
-                 use_ssl=False, max_url_len=None):
+    # TODO(eezhova): Consider changing the default host value to prevent
+    # possible binding to all interfaces. The most appropriate value seems
+    # to be 127.0.0.1, but it has to be verified that the change wouldn't
+    # break any consuming project.
+    def __init__(self, conf, name, app, host='0.0.0.0', port=0,  # nosec
+                 pool_size=None, protocol=eventlet.wsgi.HttpProtocol,
+                 backlog=128, use_ssl=False, max_url_len=None,
+                 logger_name='eventlet.wsgi.server',
+                 socket_family=None, socket_file=None, socket_mode=None):
         """Initialize, but do not start, a WSGI server.
 
+        :param conf: Instance of ConfigOpts.
         :param name: Pretty name for logging.
         :param app: The WSGI application to serve.
         :param host: IP address to serve the application.
         :param port: Port number to server the application.
         :param pool_size: Maximum number of eventlets to spawn concurrently.
+        :param protocol: Protocol class.
         :param backlog: Maximum number of queued connections.
+        :param use_ssl: Wraps the socket in an SSL context if True.
         :param max_url_len: Maximum length of permitted URLs.
+        :param logger_name: The name for the logger.
+        :param socket_family: Socket family.
+        :param socket_file: location of UNIX socket.
+        :param socket_mode: UNIX socket mode.
         :returns: None
         :raises: InvalidInput
+        :raises: EnvironmentError
         """
 
         self.conf = conf
@@ -82,15 +102,29 @@ class Server(object):
         self._protocol = protocol
         self.pool_size = pool_size or self.default_pool_size
         self._pool = eventlet.GreenPool(self.pool_size)
-        self._logger = logging.getLogger("eventlet.wsgi.server")
+        self._logger = logging.getLogger(logger_name)
         self._use_ssl = use_ssl
         self._max_url_len = max_url_len
         self.client_socket_timeout = conf.client_socket_timeout or None
-        self.default_pool_size = conf.wsgi_default_pool_size
 
         if backlog < 1:
-            raise InvalidInput(reason='The backlog must be more than 0')
+            raise InvalidInput(reason=_('The backlog must be more than 0'))
 
+        if not socket_family or socket_family in [socket.AF_INET,
+                                                  socket.AF_INET6]:
+            self.socket = self._get_socket(host, port, backlog)
+        elif hasattr(socket, "AF_UNIX") and socket_family == socket.AF_UNIX:
+            self.socket = self._get_unix_socket(socket_file, socket_mode,
+                                                backlog)
+        else:
+            raise ValueError(_("Unsupported socket family: %s"), socket_family)
+
+        (self.host, self.port) = self.socket.getsockname()[0:2]
+
+        if self._use_ssl:
+            sslutils.is_enabled(conf)
+
+    def _get_socket(self, host, port, backlog):
         bind_addr = (host, port)
         # TODO(dims): eventlet's green dns/socket module does not actually
         # support IPv6 in getaddrinfo(). We need to get around this in the
@@ -105,19 +139,25 @@ class Server(object):
         except Exception:
             family = socket.AF_INET
 
-        if self._use_ssl:
-            sslutils.is_enabled(conf)
-
         try:
-            self._socket = eventlet.listen(bind_addr, family, backlog=backlog)
+            sock = eventlet.listen(bind_addr, family, backlog=backlog)
         except EnvironmentError:
             LOG.error(_LE("Could not bind to %(host)s:%(port)s"),
                       {'host': host, 'port': port})
             raise
-
-        (self.host, self.port) = self._socket.getsockname()[0:2]
+        sock = self._set_socket_opts(sock)
         LOG.info(_LI("%(name)s listening on %(host)s:%(port)s"),
-                 {'name': self.name, 'host': self.host, 'port': self.port})
+                 {'name': self.name, 'host': host, 'port': port})
+        return sock
+
+    def _get_unix_socket(self, socket_file, socket_mode, backlog):
+        sock = eventlet.listen(socket_file, family=socket.AF_UNIX,
+                               backlog=backlog)
+        if socket_mode is not None:
+            os.chmod(socket_file, socket_mode)
+        LOG.info(_LI("%(name)s listening on %(socket_file)s:"),
+                 {'name': self.name, 'socket_file': socket_file})
+        return sock
 
     def start(self):
         """Start serving a WSGI application.
@@ -129,9 +169,7 @@ class Server(object):
         # give bad file descriptor error. So duplicating the socket object,
         # to keep file descriptor usable.
 
-        self.dup_socket = self._socket.dup()
-
-        self.dup_socket = self._set_socket_opts(self.dup_socket)
+        self.dup_socket = self.socket.dup()
 
         if self._use_ssl:
             self.dup_socket = sslutils.wrap(self.conf, self.dup_socket)
@@ -285,7 +323,7 @@ class Loader(object):
     def __init__(self, conf):
         """Initialize the loader, and attempt to find the config.
 
-        :param conf
+        :param conf: Application config
         :returns: None
 
         """
@@ -306,7 +344,7 @@ class Loader(object):
 
         :param name: Name of the application to load.
         :returns: Paste URLMap object wrapping the requested application.
-        :raises: `PasteAppNotFound`
+        :raises: PasteAppNotFound
 
         """
         try:
